import asyncio
import io

from aiosmb import logger
from aiosmb.protocol.smb.message import *
from aiosmb.protocol.smb2.message import *

class NetBIOSTransport:
	"""
	Converts incoming bytestream from the network starsport to SMB messages and vice-versa.
	This layer is presented so the network transport can be changed for a TCP/UDP/whatever type of transport.
	"""
	def __init__(self, network_transport):
		self.network_transport = network_transport
		self.socket_out_queue = network_transport.out_queue
		self.socket_in_queue = network_transport.in_queue
		
		self.in_queue = asyncio.Queue()
		self.out_queue = asyncio.Queue()
		
		self.outgoing_task = None
		self.incoming_task = None
		
	async def stop(self):
		"""
		Stops the input output processing
		"""
		if self.outgoing_task is not None:
			self.outgoing_task.cancel()
		if self.incoming_task is not None:
			self.incoming_task.cancel()
		
		
	async def run(self):
		"""
		Starts the input and output processing
		"""
		self.incoming_task = asyncio.create_task(self.handle_incoming())
		self.outgoing_task = asyncio.create_task(self.handle_outgoing())
		
	async def parse_buffer(self, buffer, total_size = None):
		"""
		Parses the incoming bytes buffer, dispatches SMBv1 or SMBv2 messages to the in_queue
		Returns a bytes array with the remaining data
		"""
		if len(buffer) > 4:
			if not total_size:
				total_size = int.from_bytes(buffer[1:4], byteorder='big', signed = False) + 4
			
			if len(buffer) >= total_size:
				msg_data = buffer[:total_size][4:]
				buffer = buffer[total_size:]
				total_size = None
					
				#print('%s nbmsg! ' % (self.network_transport.writer.get_extra_info('peername')[0], ))
				#print('[NetBIOS] MSG dispatched')
				await self.in_queue.put( (msg_data, None) )
				await self.parse_buffer(buffer, total_size)
		
		return buffer
		
	async def handle_incoming(self):
		"""
		Reads data bytes from the socket_in_queue and parses the NetBIOS messages and the SMBv1/2 messages.
		Dispatches the SMBv1/2 message objects.
		"""
		try:
			buffer = b''
			while True:
				data, err = await self.socket_in_queue.get()
				if err is not None:
					raise err
				#parse
				buffer += data
				buffer = await self.parse_buffer(buffer)

		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
			
		except Exception as e:
			logger.exception('NetBIOSTransport handle_incoming error')
			await self.in_queue.put( (None, e) )
			await self.stop()
		
	async def handle_outgoing(self):
		"""
		Reads SMBv1/2 outgoing message data bytes from out_queue, wraps them in NetBIOS object, then serializes them, then sends them to socket_out_queue
		"""
		try:
			while True:
				smb_msg_data = await self.out_queue.get()
				data  = b'\x00'
				data += len(smb_msg_data).to_bytes(3, byteorder='big', signed = False)
				data += smb_msg_data
				await self.socket_out_queue.put(data)
		
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
		
		except Exception as e:
			logger.exception('NetBIOSTransport handle_outgoing')
			await self.stop()
			



#
#
#
#
#
#


import enum
import asyncio
import ipaddress

from aiosmb import logger
from aiosmb.commons.exceptions import *
from aiosmb.network.proxies.socks5 import SOCKS5Method, SOCKS5Nego, SOCKS5NegoReply, SOCKS5Command, SOCKS5Request, SOCKS5ReplyType, SOCKS5Reply, SOCKS5PlainAuth

class Socks5ProxyConnection:
	"""
	Generic asynchronous TCP socket class, nothing SMB related.
	Creates the connection and channels incoming/outgoing bytes via asynchonous queues.
	"""
	def __init__(self, target = None, socket = None):
		self.target = target
		self.socket = socket #for future, if we want a custom soscket
		self.reader = None
		self.writer = None
		
		self.out_queue = asyncio.Queue()
		self.in_queue = asyncio.Queue()
		
		self.disconnected = asyncio.Event()

		self.incoming_task = None
		self.outgoing_task = None
		
	async def disconnect(self):
		"""
		Disconnects from the socket.
		Stops the reader and writer streams.
		"""
		if self.outgoing_task is not None:
			self.outgoing_task.cancel()
		if self.incoming_task is not None:
			self.incoming_task.cancel()

		if self.writer is not None:
			try:
				self.writer.close()
			except:
				pass
		self.writer = None
		self.reader = None
		self.disconnected.set()

	async def handle_incoming(self):
		"""
		Reads data bytes from the socket and dispatches it to the incoming queue
		"""
		try:
			timeout = int(self.target.proxy.timeout)
			while not self.disconnected.is_set():			
				data = await asyncio.gather(*[asyncio.wait_for(self.reader.read(4096), timeout)], return_exceptions = True)
				if isinstance(data[0], bytes):
					if data[0] == b'':
						await self.in_queue.put( (None, Exception('Socks5 server terminated the connection!')) )
						await self.disconnect()
					#print('%s : %s' % (self.writer.get_extra_info('peername')[0], data[0]))
					#print('SOCKS5 data in %s' % data[0])
					await self.in_queue.put( (data[0], None) )
				
				elif isinstance(data[0], asyncio.CancelledError):
					#print('SOCKS5 data in CANCELLED')
					return
					
				elif isinstance(data[0], Exception):
					#print('SOCKS5 data in exception')
					logger.exception('[SOCKS5] handle_incoming %s' % str(data[0]))
					await self.in_queue.put( (None, data[0]) )
					await self.disconnect()
					return
			
			#print('SOCKS5 data in EXITING')
		except asyncio.CancelledError:
			await self.in_queue.put( (None, asyncio.CancelledError) )
			return

		except Exception as e:
			import traceback
			traceback.print_exc()
			#print('SOCKS5 data in ERROR!')
			await self.in_queue.put( (None, e) )
			return

	async def handle_outgoing(self):
		"""
		Reads data bytes from the outgoing queue and dispatches it to the socket
		"""
		try:
			while not self.disconnected.is_set():
				data = await self.out_queue.get()
				#print('SOCKS5 data out %s' % data)
				self.writer.write(data)
				await self.writer.drain()
		except asyncio.CancelledError:
			#the SMB connection is terminating
			return
			
		except Exception as e:
			logger.exception('[SOCKS5] handle_outgoing %s' % str(e))
			await self.disconnect()
			
		
	async def connect(self):
		"""
		Main function to be called, connects to the target specified in target, and starts reading/writing.
		"""
		con = asyncio.open_connection(self.target.proxy.ip, int(self.target.proxy.port))
		try:
			self.proxy_reader, self.proxy_writer = await asyncio.wait_for(con, int(self.target.proxy.timeout))
		except asyncio.TimeoutError:
			logger.debug('[Socks5Proxy] Proxy Connection timeout')
			raise SMBConnectionTimeoutException()
			
		except ConnectionRefusedError:
			logger.debug('[Socks5Proxy] Proxy Connection refused')
			raise SMBConnectionRefusedException()
			
		except asyncio.CancelledError:
			#the SMB connection is terminating
			raise asyncio.CancelledError
			
		except Exception as e:
			logger.debug('[Socks5Proxy] connect generic exception')
			raise e

		
		#logger.info('Establishing proxy connection %s => %s' % (server.get_paddr(), target.get_paddr()))
		authmethods = [SOCKS5Method.NOAUTH]
		if self.target.proxy.username is not None:
			authmethods.append(SOCKS5Method.PLAIN)
		
		try:
			#logger.debug('Sending negotiation command to %s:%d' % proxy_writer.get_extra_info('peername'))
			self.proxy_writer.write(SOCKS5Nego.construct(authmethods).to_bytes())
			await asyncio.wait_for(self.proxy_writer.drain(), timeout = int(self.target.proxy.timeout))

			rep_nego = await asyncio.wait_for(SOCKS5NegoReply.from_streamreader(self.proxy_reader), timeout = int(self.target.proxy.timeout))
			logger.debug('Got negotiation reply from %s: %s' % (self.proxy_writer.get_extra_info('peername'), repr(rep_nego)))
			
			if rep_nego.METHOD == SOCKS5Method.PLAIN:
				logger.debug('Preforming plaintext auth to %s:%d' % self.proxy_writer.get_extra_info('peername'))
				self.proxy_writer.write(SOCKS5PlainAuth.construct(self.target.proxy.username, self.target.proxy.secret).to_bytes())
				await asyncio.wait_for(self.proxy_writer.drain(), timeout=int(self.target.proxy.timeout))
				rep_auth_nego = await asyncio.wait_for(SOCKS5NegoReply.from_streamreader(self.proxy_reader), timeout = int(self.target.proxy.timeout))

				if rep_auth_nego.METHOD != SOCKS5Method.NOAUTH:
					raise Exception('Failed to connect to proxy %s:%d! Authentication failed!' % self.proxy_writer.get_extra_info('peername'))

			logger.debug('Sending connect request to %s:%d' % self.proxy_writer.get_extra_info('peername'))
			self.proxy_writer.write(SOCKS5Request.construct(SOCKS5Command.CONNECT, self.target.get_hostname_or_ip(), int(self.target.port)).to_bytes())
			await asyncio.wait_for(self.proxy_writer.drain(), timeout=int(self.target.proxy.timeout))

			rep = await asyncio.wait_for(SOCKS5Reply.from_streamreader(self.proxy_reader), timeout=int(self.target.proxy.timeout))
			if rep.REP != SOCKS5ReplyType.SUCCEEDED:
				logger.info('Failed to connect to proxy %s! Server replied: %s' % (self.proxy_writer.get_extra_info('peername'), repr(rep.REP)))
				raise SMBSocks5ConnectionError('Socks5 remote end failed to connect to target! Reson: %s' % rep.REP.name)
			
			logger.debug('Server reply from %s : %s' % (self.proxy_writer.get_extra_info('peername'),repr(rep)))
		
		except asyncio.TimeoutError:
			logger.debug('[Socks5Proxy] Proxy Connection timeout')
			raise SMBConnectionTimeoutException()
		
		except asyncio.CancelledError:
			#the SMB connection is terminating
			raise asyncio.CancelledError
				
		except Exception as e:
			raise SMBSocks5ConnectionError('Error happened while establighing socket. Reson: %s' % e)
			logger.debug('[Socks5Proxy] connect generic exception')
			raise e
		
		else:
			logger.debug('Same socket can be used now on %s:%d' % (self.proxy_writer.get_extra_info('peername')))
			#this means that the communication can continue on the same socket!
			logger.info('Proxy connection succeeded')
			self.reader = self.proxy_reader
			self.writer = self.proxy_writer
			self.incoming_task = asyncio.create_task(self.handle_incoming())
			self.outgoing_task = asyncio.create_task(self.handle_outgoing())
			return

			

			


			


import enum


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/77b36d0f-6016-458a-a7a0-0f4a72ae1534
# File_Pipe_Printer_Access_Mask
class FileAccessMask(enum.IntFlag):
	FILE_READ_DATA = 0x00000001 #This value indicates the right to read data from the file or named pipe.
	FILE_WRITE_DATA = 0x00000002 #This value indicates the right to write data into the file or named pipe beyond the end of the file.
	FILE_APPEND_DATA = 0x00000004 #This value indicates the right to append data into the file or named pipe.
	FILE_READ_EA = 0x00000008 #This value indicates the right to read the extended attributes of the file or named pipe.
	FILE_WRITE_EA = 0x00000010 #This value indicates the right to write or change the extended attributes to the file or named pipe.
	FILE_DELETE_CHILD = 0x00000040 #This value indicates the right to delete entries within a directory.
	FILE_EXECUTE = 0x00000020 #This value indicates the right to execute the file.
	FILE_READ_ATTRIBUTES = 0x00000080 #This value indicates the right to read the attributes of the file.
	FILE_WRITE_ATTRIBUTES = 0x00000100 #This value indicates the right to change the attributes of the file.
	DELETE = 0x00010000 #This value indicates the right to delete the file.
	READ_CONTROL = 0x00020000 #This value indicates the right to read the security descriptor for the file or named pipe.
	WRITE_DAC = 0x00040000 #This value indicates the right to change the discretionary access control list (DACL) in the security descriptor for the file or named pipe. For the DACL data structure, see ACL in [MS-DTYP].
	WRITE_OWNER = 0x00080000 #This value indicates the right to change the owner in the security descriptor for the file or named pipe.
	SYNCHRONIZE = 0x00100000 #SMB2 clients set this flag to any value.<40> #SMB2 servers SHOULD<41> ignore this flag.
	ACCESS_SYSTEM_SECURITY = 0x01000000 #This value indicates the right to read or change the system access control list (SACL) in the security descriptor for the file or named pipe. For the SACL data structure, see ACL in [MS-DTYP].<42>
	MAXIMUM_ALLOWED = 0x02000000 #This value indicates that the client is requesting an open to the file with the highest level of access the client has on this file. If no access is granted for the client on this file, the server MUST fail the open with STATUS_ACCESS_DENIED.
	GENERIC_ALL = 0x10000000 #This value indicates a request for all the access flags that are previously listed except MAXIMUM_ALLOWED and ACCESS_SYSTEM_SECURITY.
	GENERIC_EXECUTE = 0x20000000 #This value indicates a request for the following combination of access flags listed above: FILE_READ_ATTRIBUTES| FILE_EXECUTE| SYNCHRONIZE| READ_CONTROL.
	GENERIC_WRITE = 0x40000000 #This value indicates a request for the following combination of access flags listed above: FILE_WRITE_DATA| FILE_APPEND_DATA| FILE_WRITE_ATTRIBUTES| FILE_WRITE_EA| SYNCHRONIZE| READ_CONTROL.
	GENERIC_READ = 0x80000000 #This value indicates a request for the following combination of access flags listed above: FILE_READ_DATA| FILE_READ_ATTRIBUTES| FILE_READ_EA| SYNCHRONIZE| READ_CONTROL.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/0a5934b1-80f1-4da0-b1bf-5e021c309b71
class DirectoryAccessMask(enum.IntFlag):
	FILE_LIST_DIRECTORY = 0x00000001 #This value indicates the right to enumerate the contents of the directory.
	FILE_ADD_FILE = 0x00000002 #This value indicates the right to create a file under the directory.
	FILE_ADD_SUBDIRECTORY = 0x00000004 #This value indicates the right to add a sub-directory under the directory.
	FILE_READ_EA = 0x00000008 #This value indicates the right to read the extended attributes of the directory.
	FILE_WRITE_EA = 0x00000010 #This value indicates the right to write or change the extended attributes of the directory.
	FILE_TRAVERSE = 0x00000020 #This value indicates the right to traverse this directory if the server enforces traversal checking.
	FILE_DELETE_CHILD = 0x00000040 #This value indicates the right to delete the files and directories within this directory.
	FILE_READ_ATTRIBUTES = 0x00000080 #This value indicates the right to read the attributes of the directory.
	FILE_WRITE_ATTRIBUTES = 0x00000100 #This value indicates the right to change the attributes of the directory.
	DELETE = 0x00010000 #This value indicates the right to delete the directory.
	READ_CONTROL = 0x00020000 #This value indicates the right to read the security descriptor for the directory.
	WRITE_DAC = 0x00040000 #This value indicates the right to change the DACL in the security descriptor for the directory. For the DACL data structure, see ACL in [MS-DTYP].
	WRITE_OWNER = 0x00080000 #This value indicates the right to change the owner in the security descriptor for the directory.
	SYNCHRONIZE = 0x00100000 #SMB2 clients set this flag to any value.<43> SMB2 servers SHOULD<44> ignore this flag.
	ACCESS_SYSTEM_SECURITY = 0x01000000 #This value indicates the right to read or change the SACL in the security descriptor for the directory. For the SACL data structure, see ACL in [MS-DTYP].<45>
	MAXIMUM_ALLOWED = 0x02000000 #This value indicates that the client is requesting an open to the directory with the highest level of access the client has on this directory. If no access is granted for the client on this directory, the server MUST fail the open with STATUS_ACCESS_DENIED.
	GENERIC_ALL = 0x10000000 #This value indicates a request for all the access flags that are listed above except MAXIMUM_ALLOWED and ACCESS_SYSTEM_SECURITY.
	GENERIC_EXECUTE = 0x20000000 #This value indicates a request for the following access flags listed above: FILE_READ_ATTRIBUTES| FILE_TRAVERSE| SYNCHRONIZE| READ_CONTROL.
	GENERIC_WRITE = 0x40000000 #This value indicates a request for the following access flags listed above: FILE_ADD_FILE| FILE_ADD_SUBDIRECTORY| FILE_WRITE_ATTRIBUTES| FILE_WRITE_EA| SYNCHRONIZE| READ_CONTROL.
	GENERIC_READ = 0x80000000 #This value indicates a request for the following access flags listed above: FILE_LIST_DIRECTORY| FILE_READ_ATTRIBUTES| FILE_READ_EA| SYNCHRONIZE| READ_CONTROL.



from aiosmb.protocol.smb.commands.negotiate_req import SMB_COM_NEGOTIATE_REQ

#add new commands as they are implemented.
#do not forget to add the same commands to command_cdoes as well!


__all__ = [
	'SMB_COM_NEGOTIATE_REQ',

]


import io
import enum

from aiosmb.wintypes.access_mask import *

class ReadFlag(enum.IntFlag):
	SMB2_READFLAG_READ_UNBUFFERED = 0x01 #The data is read directly from the underlying storage.
	SMB2_READFLAG_REQUEST_COMPRESSED = 0x02 #The server is requested to compress the read response when responding to the request. This flag is not valid for the SMB 2.0.2, 2.1, 3.0 and 3.0.2 dialects<53>.

class Channel(enum.Enum):
	SMB2_CHANNEL_NONE = 0x00000000 #No channel information is present in the request. The ReadChannelInfoOffset and ReadChannelInfoLength fields MUST be set to 0 by the client and MUST be ignored by the server.
	SMB2_CHANNEL_RDMA_V1 = 0x00000001 #One or more SMB_DIRECT_BUFFER_DESCRIPTOR_V1 structures as specified in [MS-SMBD] section 2.2.3.1 are present in the channel information specified by ReadChannelInfoOffset and ReadChannelInfoLength fields.
	SMB2_CHANNEL_RDMA_V1_INVALIDATE = 0x00000002 #This flag is not valid for the SMB 3.0 dialect. One or more SMB_DIRECT_BUFFER_DESCRIPTOR_V1 structures, as specified in [MS-SMBD] section 2.2.3.1, are present in the channel information specified by the ReadChannelInfoOffset and ReadChannelInfoLength fields. The server is requested to perform remote invalidation when responding to the request as specified in [MS-SMBD] section 3.1.4.2.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/320f04f3-1b28-45cd-aaa1-9e5aed810dca
class READ_REQ:
	def __init__(self):
		self.StructureSize = 49
		self.Padding = 0x50 #dunno why :(
		self.Flags = 0
		self.Length = None
		self.Offset = None
		self.FileId = None
		self.MinimumCount = None
		self.Channel = Channel.SMB2_CHANNEL_NONE
		self.RemainingBytes = None
		self.ReadChannelInfoOffset = 0
		self.ReadChannelInfoLength = 0
		self.Buffer = None
		
		#high-level variable, not part of the spec!
		#For SMB3 only
		self.ReadChannelInfo = None

	def to_bytes(self):
		if self.ReadChannelInfo:
			self.Buffer = self.ReadChannelInfo.to_bytes()
			self.ReadChannelInfoOffset = 64 + 1+1+4+8+16+4+4+4+2+2
			self.ReadChannelInfoLength = len(self.ReadChannelInfo.to_bytes())
		
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Padding.to_bytes(1, byteorder='little', signed = False)
		t += self.Flags.to_bytes(1, byteorder='little', signed = False)
		t += self.Length.to_bytes(4, byteorder='little', signed = False)
		t += self.Offset.to_bytes(8, byteorder='little', signed = False)
		t += self.FileId.to_bytes(16, byteorder='little', signed = False)
		t += self.MinimumCount.to_bytes(4, byteorder='little', signed = False)
		t += self.Channel.value.to_bytes(4, byteorder='little', signed = False)
		t += self.RemainingBytes.to_bytes(4, byteorder='little', signed = False)
		t += self.ReadChannelInfoOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.ReadChannelInfoLength.to_bytes(2, byteorder='little', signed = False)
		if self.ReadChannelInfoOffset > 0:
			t += self.Buffer
		else:
			t += b'\x00' #dont ask me why...
		return t

	@staticmethod
	def from_bytes(bbuff):
		return READ_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = READ_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 49
		msg.Padding = int.from_bytes(buff.read(1), byteorder='little')
		msg.Flags = ReadFlag(int.from_bytes(buff.read(1), byteorder='little'))
		msg.Length = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.Offset = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.FileId = int.from_bytes(buff.read(16), byteorder = 'little')
		msg.MinimumCount = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.Channel = Channel(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.RemainingBytes = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.ReadChannelInfoOffset = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.ReadChannelInfoLength = int.from_bytes(buff.read(2), byteorder = 'little')

		if msg.ReadChannelInfoOffset > 0:
			buff.seek(msg.ReadChannelInfoOffset, io.SEEK_SET)
			msg.Buffer= buff.read(msg.ReadChannelInfoLength)
			
			#TODO: correct parsing in SMB3!
			msg.ReadChannelInfo = msg.Buffer.read()
		return msg

	def __repr__(self):
		t = '==== SMB2 READ REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'Padding: %s\r\n' % self.Padding
		t += 'Flags: %s\r\n' % self.Flags
		t += 'Length: %s\r\n' % self.Length
		t += 'Offset: %s\r\n' % self.Offset
		t += 'FileId: %s\r\n' % self.FileId
		t += 'MinimumCount: %s\r\n' % self.MinimumCount
		t += 'Channel: %s\r\n' % self.Channel
		t += 'RemainingBytes: %s\r\n' % self.RemainingBytes
		t += 'ReadChannelInfoOffset: %s\r\n' % self.ReadChannelInfoOffset
		t += 'ReadChannelInfoLength: %s\r\n' % self.ReadChannelInfoLength
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/3e3d2f2c-0e2f-41ea-ad07-fbca6ffdfd90
class READ_REPLY:
	def __init__(self):
		self.StructureSize = 17
		self.DataOffset = None
		self.Reserved = 0
		self.DataLength = None
		self.DataRemaining = None
		self.Reserved2 = 0
		self.Buffer  = None

	def to_bytes(self):
		#todo
		pass

	@staticmethod
	def from_bytes(bbuff):
		return READ_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = READ_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 17
		msg.DataOffset = int.from_bytes(buff.read(1), byteorder='little')
		msg.Reserved  = int.from_bytes(buff.read(1), byteorder='little')
		msg.DataLength = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.DataRemaining = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.Reserved2 = int.from_bytes(buff.read(4), byteorder = 'little')
		
		buff.seek(msg.DataOffset, io.SEEK_SET)
		msg.Buffer= buff.read(msg.DataLength)
		return msg

	def __repr__(self):
		t = '==== SMB2 READ REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'DataOffset: %s\r\n' % self.DataOffset
		t += 'DataLength: %s\r\n' % self.DataLength
		t += 'DataRemaining: %s\r\n' % self.DataRemaining
		t += 'Buffer: %s\r\n' % self.Buffer
		return t


from aiosmb.dcerpc.v5.common.connection.authentication import DCERPCAuth
from aiosmb.dcerpc.v5.common.connection.target import DCERPCSMBTarget
from aiosmb.dcerpc.v5.connection import DCERPC5Connection
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_GSS_NEGOTIATE, RPC_C_AUTHN_WINNT

class SMBDCEFactory:
	def __init__(self, smb_connection, filename):
		self.smb_connection = smb_connection
		self.filename = filename
	
	def get_dce_rpc(self):
		dcerpc_target = DCERPCSMBTarget(
			None, 
			self.smb_connection.target.get_hostname_or_ip(), 
			pipe = self.filename, 
			smb_connection = self.smb_connection, 
			timeout = self.smb_connection.target.timeout
		)
		#print(str(dcerpc_target))
		dcerpc_auth = DCERPCAuth.from_smb_gssapi(self.smb_connection.gssapi)
		#print(str(dcerpc_auth))
		conn = DCERPC5Connection(dcerpc_auth, dcerpc_target)
		
		if dcerpc_auth.ntlm is not None:
			conn.set_auth_type(RPC_C_AUTHN_WINNT)
		if dcerpc_auth.kerberos is not None:
			conn.set_auth_type(RPC_C_AUTHN_GSS_NEGOTIATE)

		#auth level must be set elsewhere, this function cannot fuigure out your needs!
		
		return conn


"""
The idea here is to offer compatibility with 3rd party libraries by extending wrappers for ech encryption mode
This is needed because the pure python implementation for encryption and hashing algorithms are quite slow

currently it's not the perfect wrapper, needs to be extended
"""

from aiosmb.crypto.BASE import symmetricBASE, cipherMODE
import aiosmb.crypto.pure.DES.DES as _pyDES
try:
	from Crypto.Cipher import DES as _pyCryptoDES
except:
	pass

try:
	from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
	from cryptography.hazmat.backends import default_backend
except:
	pass


# from impacket
def expand_DES_key(key):
	# Expand the key from a 7-byte password key into a 8-byte DES key
	key  = key[:7]
	key += b'\x00'*(7-len(key))
	s  = (((key[0] >> 1) & 0x7f) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[0] & 0x01) << 6 | ((key[1] >> 2) & 0x3f)) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[1] & 0x03) << 5 | ((key[2] >> 3) & 0x1f)) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[2] & 0x07) << 4 | ((key[3] >> 4) & 0x0f)) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[3] & 0x0f) << 3 | ((key[4] >> 5) & 0x07)) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[4] & 0x1f) << 2 | ((key[5] >> 6) & 0x03)) << 1).to_bytes(1, byteorder = 'big')
	s += (((key[5] & 0x3f) << 1 | ((key[6] >> 7) & 0x01)) << 1).to_bytes(1, byteorder = 'big')
	s += ( (key[6] & 0x7f) << 1).to_bytes(1, byteorder = 'big')
	return s
#

class pureDES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None):
		self.key = key
		if len(key) == 7:
			self.key = expand_DES_key(key)

		self.mode = mode
		self.IV = IV
		symmetricBASE.__init__(self)

	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			mode = _pyDES.ECB
		elif self.mode == cipherMODE.CBC:
			mode = _pyDES.CBC
		else:
			raise Exception('Unknown cipher mode!')
		
		self._cipher = _pyDES.des(self.key, mode, self.IV)

	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)



class pyCryptoDES(symmetricBASE):
	def __init__(self, key, mode = cipherMODE.ECB, IV = None):
		self.key = key
		if len(key) == 7:
			self.key = __expand_DES_key(key)

		self.mode = mode
		self.IV = IV
		symmetricBASE.__init__(self)

	def setup_cipher(self):
		if self.mode == cipherMODE.ECB:
			self._cipher = _pyCryptoDES.new(self.key)
		elif self.mode == cipherMODE.CBC:
			self._cipher = _pyCryptoDES.new(self.key, _pyCryptoDES.MODE_CBC, self.IV)
		else:
			raise Exception('Unknown cipher mode!')
		
		

	def encrypt(self, data):
		return self._cipher.encrypt(data)
	def decrypt(self, data):
		return self._cipher.decrypt(data)




import io
import enum

from aiosmb.wintypes.access_mask import *
from aiosmb.protocol.smb2.commands.read import Channel

class WriteFlag(enum.IntFlag):
	NONE = 0
	WRITE_THROUGH = 0x00000001 #The server performs File write-through on the write operation. This value is not valid for the SMB 2.0.2 dialect.
	WRITE_UNBUFFERED = 0x00000002 #File buffering is not performed. This bit is not valid for the SMB 2.0.2, 2.1, and 3.0 dialects.


# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/e7046961-3318-4350-be2a-a8d69bb59ce8
class WRITE_REQ:
	def __init__(self):
		self.StructureSize = 49
		self.DataOffset = None
		self.Length = None
		self.Offset = None
		self.FileId = None
		self.Channel = Channel.SMB2_CHANNEL_NONE
		self.RemainingBytes = 0
		self.WriteChannelInfoOffset = 0
		self.WriteChannelInfoLength = 0
		self.Flags = WriteFlag.NONE
		self.Buffer = None
		
		#high-level variable, not part of the spec!
		#For SMB3 only
		self.WriteChannelInfo = None
		self.Data = None

	def to_bytes(self):
		self.DataOffset = 64 + 2+2+4+8+16+4+4+2+2+4
		if self.Data:
			self.Length = len(self.Data)
			self.Buffer = self.Data
		else:
			self.Length = 0 
			self.Buffer = b''
			
		if self.WriteChannelInfo:
			self.Buffer += self.WriteChannelInfo.to_bytes()
			self.WriteChannelInfoOffset = self.DataOffset + len(self.Buffer)
			self.WriteChannelInfoLength = len(self.WriteChannelInfo.to_bytes())
		
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.DataOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.Length.to_bytes(4, byteorder='little', signed = False)
		t += self.Offset.to_bytes(8, byteorder='little', signed = False)
		t += self.FileId.to_bytes(16, byteorder='little', signed = False)
		t += self.Channel.value.to_bytes(4, byteorder='little', signed = False)
		t += self.RemainingBytes.to_bytes(4, byteorder='little', signed = False)
		t += self.WriteChannelInfoOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.WriteChannelInfoLength.to_bytes(2, byteorder='little', signed = False)
		t += self.Flags.value.to_bytes(4, byteorder='little', signed = False)
		
		t += self.Buffer
		return t

	@staticmethod
	def from_bytes(bbuff):
		return WRITE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = WRITE_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 49
		msg.DataOffset = int.from_bytes(buff.read(2), byteorder='little')
		msg.Length  = int.from_bytes(buff.read(4), byteorder='little')
		msg.Offset = int.from_bytes(buff.read(8), byteorder = 'little')
		msg.FileId = int.from_bytes(buff.read(16), byteorder = 'little')
		msg.Channel = Channel(int.from_bytes(buff.read(4), byteorder = 'little'))
		msg.RemainingBytes = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.WriteChannelInfoOffset  = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.WriteChannelInfoLength  = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.Flags  = WriteFlag(int.from_bytes(buff.read(4), byteorder = 'little'))
		
		buff.seek(msg.DataOffset, io.SEEK_SET)
		msg.Data = buff.read(msg.Length)

		if msg.WriteChannelInfoOffset > 0:
			#TODO: correct parsing in SMB3!
			buff.seek(msg.WriteChannelInfoOffset, io.SEEK_SET)
			msg.WriteChannelInfo = buff.read(msg.WriteChannelInfoLength)
		
		return msg

	def __repr__(self):
		t = '==== SMB2 WRITE REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'DataOffset: %s\r\n' % self.DataOffset
		t += 'Length: %s\r\n' % self.Length
		t += 'Offset: %s\r\n' % self.Offset
		t += 'FileId: %s\r\n' % self.FileId
		t += 'Channel: %s\r\n' % self.Channel
		t += 'RemainingBytes: %s\r\n' % self.RemainingBytes
		t += 'WriteChannelInfoOffset: %s\r\n' % self.WriteChannelInfoOffset
		t += 'WriteChannelInfoLength: %s\r\n' % self.WriteChannelInfoLength
		t += 'Flags: %s\r\n' % self.Flags
		t += 'Data: %s\r\n' % self.Data
		t += 'WriteChannelInfo: %s\r\n' % self.WriteChannelInfo
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/7b80a339-f4d3-4575-8ce2-70a06f24f133
class WRITE_REPLY:
	def __init__(self):
		self.StructureSize = 17
		self.Reserved = 0
		self.Count = None
		self.Remaining = None
		self.WriteChannelInfoOffset = 0
		self.WriteChannelInfoLength = 0

	def to_bytes(self):
		#todo
		pass

	@staticmethod
	def from_bytes(bbuff):
		return WRITE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = WRITE_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 17
		msg.Reserved  = int.from_bytes(buff.read(2), byteorder='little')
		msg.Count  = int.from_bytes(buff.read(4), byteorder='little')
		msg.Remaining  = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.WriteChannelInfoOffset  = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.WriteChannelInfoLength  = int.from_bytes(buff.read(2), byteorder = 'little')
		
		return msg

	def __repr__(self):
		t = '==== SMB2 WRITE REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'Count: %s\r\n' % self.Count
		t += 'Remaining: %s\r\n' % self.Remaining
		t += 'WriteChannelInfoOffset: %s\r\n' % self.WriteChannelInfoOffset
		t += 'WriteChannelInfoLength: %s\r\n' % self.WriteChannelInfoLength
		return t


import asyncio
import logging
import json

import aiosmb
from aiosmb.commons.connection.credential import SMBCredential
from aiosmb.commons.connection.target import SMBTarget
from aiosmb.commons.connection.authbuilder import AuthenticatorBuilder
from aiosmb.connection import SMBConnection

from aiosmb.dcerpc.v5.transport.smbtransport import SMBTransport
from aiosmb.dcerpc.v5.interfaces.samrmgr import SMBSAMR
from aiosmb.dcerpc.v5 import samr

class SMBDomain:
	def __init__(self, domain_name, samr = None, connection = None):
		self.domain_name = domain_name
		self.connection = connection
		self.domain_access_level = None
		self.samr = samr
		self.domain_sid = None
		self.domain_handle = None
		
		if samr is None and connection is None:
			raise Exception('Either SAMR or CONNECTION must be provided!')
			
	async def __aenter__(self):
		return self
		
	async def __aexit__(self, exc_type, exc, traceback):
		pass
	
	async def open(self, access_level = samr.MAXIMUM_ALLOWED):
		self.domain_access_level = access_level
		if self.samr is None:
			self.samr = SMBSAMR(self.connection)
			logging.debug('Connecting to SAMR')
			try:
				await self.samr.connect()
			except Exception as e:
				logging.exception('Failed to connect to SAMR')
				raise e
				
		self.domain_sid = await self.samr.get_domain_sid(self.domain_name)
		self.domain_handle = await self.samr.open_domain(self.domain_sid, access_level = self.domain_access_level)
		#print(self.domain_sid)
		
	async def get_info(self):
		for i in [
				samr.DOMAIN_INFORMATION_CLASS.DomainPasswordInformation,
				samr.DOMAIN_INFORMATION_CLASS.DomainGeneralInformation2,
			]:
			info = await self.samr.get_info(self.domain_handle, i)
			info.dump()	
		
		return info
		
	async def list_users(self):
		async for user_name, user_sid in self.samr.enumerate_users(self.domain_handle):
			yield (user_name, user_sid)
			
	async def open_user(self, user_sid, access_level = samr.MAXIMUM_ALLOWED):
		user_rid = user_sid.replace(self.domain_sid, '')[1:]
		user_handle = await self.samr.open_user(self.domain_handle, int(user_rid), access_level = access_level)
		return user_handle
		
	async def get_user_group_memberships(self, user_handle):
		async for group_sid in self.samr.get_user_group_memberships(user_handle):
			yield group_sid
			
	async def get_user_info(self, user_handle):
		info = await self.samr.get_user_info(user_handle, samr.USER_INFORMATION_CLASS.UserParametersInformation)
		info.dump()
		return info
		
	async def list_groups(self):
		async for name, sid in self.samr.list_domain_groups(self.domain_handle):
			yield name, sid
			
	async def get_security_info(self, handle):
		info = await self.samr.get_security_info(handle)
		info.dump()
		return info
	
		
	
	
	
	
async def domain_test(connection_string, domain_name, out_file = None, json_out = False):
	target = SMBTarget.from_connection_string(connection_string)
	credential = SMBCredential.from_connection_string(connection_string)
	spneg = AuthenticatorBuilder.to_spnego_cred(credential, target)
	
	async with SMBConnection(spneg, target) as connection: 
		await connection.login()
		
		async with SMBDomain(domain_name, connection=connection) as domain:
			logging.debug('Connecting to SAMR')
			try:
				await domain.open(access_level = samr.DOMAIN_ALL_ACCESS)
			except Exception as e:
				logging.exception('Failed to create domain object')
				
			info = await domain.get_info()
			
			async for user_name, user_sid in domain.list_users():
				print(user_name, user_sid)
				try:
					user_handle = await domain.open_user(user_sid, access_level = samr.USER_ALL_ACCESS)
					#x = await domain.get_security_info(user_handle)
					user_info = await domain.get_user_info(user_handle)
					
					#async for group_sid in domain.get_user_group_memberships(user_handle):
					#	print(group_sid)
				except Exception as e:
					print(e)
					continue
					
			#async for name, sid in domain.list_groups():
			#	print(name, sid)

	print('Done!')
if __name__ == '__main__':
	import argparse
	
	parser = argparse.ArgumentParser(description='List user sessions on a remote machine')
	parser.add_argument('-v', '--verbose', action='count', default=0, help='Increase verbosity, can be stacked')
	parser.add_argument('connection_string', help='connection string. Identifies the credential to be used and the target')
	parser.add_argument('-o', '--out-file', help='file to store the results in')
	parser.add_argument('--json', action='store_true', help='File output will be pritten in JSON format')
	args = parser.parse_args()	
	
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
		aiosmb.logger.setLevel(logging.WARNING)
		
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
		aiosmb.logger.setLevel(logging.INFO)
		
	else:
		logging.basicConfig(level=1)
		aiosmb.logger.setLevel(logging.DEBUG)
	
	asyncio.run(domain_test(args.connection_string, 'TEST', out_file = args.out_file, json_out = args.json))
	